#!/usr/bin/perl
use lib "./UPnP/lib";
use lib ".";
use UPnP::ControlPoint;
use Socket;
use IO::Select;
use IO::Handle;
use Data::Dumper;
use HTML::Parser;
use HTML::Entities;
use URI::Escape;
use XML::Simple;
use HTTP::Daemon;
use HTML::Template;
use LWP::MediaTypes qw(add_type);
use POSIX qw(strftime);
use Encode qw(encode decode);
use LWP::UserAgent;
use strict;

###############################################################################
#CONFIG
$main::MAX_LOG_LEVEL  = 2;    # Lower, the less output
$main::HTTP_PORT      = 8001; # Port our fake http server listens on
$main::MAX_SEARCH     = 400;  # Default max search results to return
$main::RENEW_SUB_TIME = 3600; # How often do we do a UPnP renew in seconds
$main::VERSION        = "0.25";
$main::DEFAULT        = "index.html";
#$main::DEFAULT        = "simple/index.html";
$main::AACACHE        = 3600; # How long do we tell browser to cache album art in secs
$| = 1;


use SOAP::Lite maptype => {}; 

$SIG{INT} = "main::quit";
$SIG{PIPE} = sub {};
$SIG{CHLD} = sub {};

$Data::Dumper::Indent  = 1;
@main::TIMERS          = ();
$main::SONOS_UPDATENUM = time();
$main::HTTP_HANDLERS   = {};
$main::HTTP_COUNTER    = 0;
$main::PREFS           = {};
$main::ZONESUPDATE     = 0;

foreach my $arg (@ARGV) {
    $main::MAX_LOG_LEVEL = 4 if ($arg eq "-debug");
}

main();

###############################################################################
sub quit {
    plugin_quit();
    http_quit();
    sonos_quit();
    Log (0, "Shutting Down");
    exit 0;
}

###############################################################################
# main
sub main {
    Log (0, "Starting up!");
    Log (0, "If the application doesn't seem to work:");
    Log (0, "  * you may need to disable your firewall or allow the application");
    Log (0, "  * make sure the computer is on the same network as Sonos boxes");
    Log (0, "  * sonos.pl currently does not work well with VPNs or multiple interfaces");

    add_type("text/css" => qw(css));
    $main::useragent = LWP::UserAgent->new(env_proxy  => 1, keep_alive => 2, parse_head => 0);
    $main::daemon = HTTP::Daemon->new(LocalPort => $main::HTTP_PORT, Reuse => 1) || die;
    $main::cp = UPnP::ControlPoint->new ();
    my $search = $main::cp->searchByType("urn:schemas-upnp-org:device:ZonePlayer:1", \&main::upnp_search_cb);

    my @selsockets = $main::cp->sockets();
    @selsockets = (@selsockets, $main::daemon);
    my $select = IO::Select->new(@selsockets);

    http_register_handler("/getaa", \&http_albumart_request);
    http_register_handler("/getAA", \&http_albumart_request);

    sonos_musicdb_load();
    sonos_prefsdb_load();
    sonos_renew_subscriptions();
    plugin_load();

    # MAIN LOOP
    while (1) {

        # Check the callbacks we have waiting
        my $timeout = 5;
        my $now = time;
        while ($#main::TIMERS >= 0) {
            if ($main::TIMERS[0][0] <= $now) {
                my($time, $callback, @args) = @{shift @main::TIMERS};
                &$callback(@args);
            } else {
                $timeout = $main::TIMERS[0][0] - $now;
                last;
            }
        }

        # Find if any sockets are ready for reading
        my @sockets = $select->can_read($timeout);

        # Call the handlers for the sockets
        for my $sock (@sockets) {
            if ($sock == $main::daemon) {
                my $c = $main::daemon->accept;
                my $r = $c->get_request;
                http_handle_request($c, $r);
            } else {
                $main::cp->handleOnce($sock);
            }
        }
    }
}

###############################################################################
# PLUGINS
###############################################################################

###############################################################################
sub plugin_load {
    Log (0, "Loading Plugins");

    opendir(DIR, "Plugins") || return;

    for my $plugin (readdir(DIR)) {
        if ($plugin =~ /(.+)\.pm$/) {
            $main::PLUGINS{$plugin} = {} if (!$main::PLUGINS{$plugin});
            $main::PLUGINS{$plugin}->{require} = "Plugins::" . $1;

        } elsif (-d "Plugins/$plugin" && -e "Plugins/$plugin/Plugin.pm") {
            $main::PLUGINS{$plugin} = {} if (!$main::PLUGINS{$plugin});
            $main::PLUGINS{$plugin}->{require} = "Plugins::" . $plugin . "::Plugin";
            if ( -d "Plugins/$plugin/html") {
                $main::PLUGINS{$plugin}->{html} = "Plugins/$plugin/html/";
            }

        }
    }
    closedir(DIR);

    foreach my $plugin (keys %main::PLUGINS) {
	eval "require ". $main::PLUGINS{$plugin}->{require};
	if ($@) {
            Log(0, "Can't load $plugin: " . $@);
            delete $main::PLUGINS{$plugin};
            next;
        }
        eval "Plugins::${plugin}::init();";
	if ($@) {
            Log(0, "Can't init $plugin: " . $@);
            delete $main::PLUGINS{$plugin};
            next;
        }
    }
}

###############################################################################
sub plugin_register {
    my ($plugin, $name, $link, $tmplhook) = @_;

    $main::PLUGINS{$plugin}->{name} = $name;
    $main::PLUGINS{$plugin}->{link} = $link;
    $main::PLUGINS{$plugin}->{tmplhook} = $tmplhook;
}

###############################################################################
sub plugin_quit {
    foreach my $plugin (keys %main::PLUGINS) {
        eval "Plugins::${plugin}::quit();";
	if ($@) {
            Log(0, "Can't quit $plugin: " . $@);
        }
    }
}

###############################################################################
# SONOS
###############################################################################

###############################################################################
sub sonos_quit {
    foreach my $sub (keys %main::SUBSCRIPTIONS) {
        Log (2, "Unsubscribe $sub");
        $main::SUBSCRIPTIONS{$sub}->unsubscribe;
    }
}
###############################################################################
sub sonos_fetch_radio_stations {
# Unlike music library, this shouldn't be huge

    my ($zone) = split(",", $main::UPDATEID{MasterRadioUpdateID});
    my $folders =  upnp_content_dir_browse($zone, "R:");

    if (! defined $folders) {
        Log(1, "$zone not available, retrying");
        add_timeout (time()+5, \&sonos_fetch_radio_stations);
        return
    }
    delete $main::MUSIC{"#Radio"};

    foreach my $folder (@{$folders}) {
        my $stations =  upnp_content_dir_browse($zone, $folder->{"id"});
        if ($stations && @{$stations}) {
            foreach my $station (@{$stations}) {
                $main::MUSIC{"#Radio"}{$folder->{"dc:title"}}{$station->{"dc:title"}} = $station;
            }
            $main::LASTUPDATE  = $main::SONOS_UPDATENUM;
            $main::MUSICUPDATE = $main::SONOS_UPDATENUM++;
        }
    }

    sonos_process_waiting("RADIO");
}

###############################################################################
sub sonos_fetch_linein {
    my ($onezone) = @_;

    if (defined $onezone) {
        foreach my $linein (keys %{$main::MUSIC{"#LineIn"}}) {
            if ($main::MUSIC{"#LineIn"}->{$linein}->{zone} eq $onezone) {
                delete $main::MUSIC{"#LineIn"}->{$linein};
                last;
            }
        }
    } else {
        delete $main::MUSIC{"#LineIn"};
    }

    foreach my $zone (keys %main::ZONES) {
        next if (defined $onezone && $onezone ne $zone);
        my $linein =  upnp_content_dir_browse($zone, "AI:");
        $linein = $linein->[0];
        $linein->{zone} = $zone;
        my $name = $main::ZONES{$zone}->{ZoneName} . " - " . $linein->{"dc:title"};
        $main::MUSIC{"#LineIn"}{$name} = $linein
    }

    sonos_process_waiting("LINEIN");
}

###############################################################################
sub sonos_fetch_playlists {
# Unlike music library, this shouldn't be huge

    my ($zone) = split(",", $main::UPDATEID{SavedQueuesUpdateID});

    my $playlists =  upnp_content_dir_browse($zone, "SQ:");
    if (! defined $playlists) {
        Log(1, "$zone not available, retrying");
        add_timeout (time()+5, \&sonos_fetch_playlists);
        return
    }
    delete $main::MUSIC{"#Playlists"};

    return if (!defined $playlists);
    foreach my $playlist (@{$playlists}) {
        $main::MUSIC{"#Playlists"}{$playlist->{"dc:title"}} = $playlist;
    }
    $main::LASTUPDATE  = $main::SONOS_UPDATENUM;
    $main::MUSICUPDATE = $main::SONOS_UPDATENUM++;

    sonos_process_waiting("PLAYLISTS");
}

###############################################################################
sub sonos_reindex {
    if ($main::UPDATEID{ShareIndexInProgress} || $main::UPDATEID{ShareIndexInProgress2}) {
        Log (1, "Alreadying reindexing");
        return;
    }
    upnp_content_dir_refresh_share_index();
}
###############################################################################
sub sonos_fetch_music {
    $main::UPDATEID{ShareIndexInProgress2} = 1;

    my ($zone) = split(",", $main::UPDATEID{ShareListUpdateID});
    my $contentDir = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:ContentDirectory:1");

    if (! defined $contentDir) {
        Log(1, "$zone not available, will retry");
        add_timeout (time()+5, \&sonos_fetch_music);
        return
    }

    $main::REINDEX{contentDirProxy} = $contentDir->controlProxy; 
    $main::REINDEX{start} = 0;
    undef %main::REINDEX_DATA;
    Log (1, "Fetching Music DB");
    sonos_fetch_music_callback();
}

###############################################################################
sub sonos_fetch_music_callback {
    # We only do 200 at a time otherwise we lock out everything else
    my $result = $main::REINDEX{contentDirProxy}->Browse("A:TRACKS", 'BrowseDirectChildren', 
                                          'dc:title,res,dc:creator,upnp:artist,upnp:album', 
                                          $main::REINDEX{start}, 200, "");

    $main::REINDEX{start} += $result->getValue("NumberReturned");
    Log (1, "DB: $main::REINDEX{start} of " . $result->getValue("TotalMatches"));

    my $results = $result->getValue("Result");
    my $tree = XMLin($results, forcearray => ["item"]);

    foreach my $objectid (keys %{$tree->{item}}) {
        my $entry = $tree->{item}{$objectid};
        my $artist = $entry->{"dc:creator"};
        my $album = $entry->{"upnp:album"};
        my $title = $entry->{"dc:title"};
        $artist = "Unknown" if (! $entry->{"dc:creator"});
        $album  = "Unknown" if (! $entry->{"upnp:album"});

# Change slash to dash
        $artist =~ s/\//-/g;
        $album  =~ s/\//-/g;
        $title  =~ s/\//-/g;

#Only copy the stuff we want
        $main::REINDEX_DATA{$artist}{$album}{$title}{"upnp:genre"} = $entry->{"upnp:genre"};
        $main::REINDEX_DATA{$artist}{$album}{$title}{"res"}{"content"} = $entry->{"res"}->{"content"};
        $main::REINDEX_DATA{$artist}{$album}{$title}{"upnp:class"} = $entry->{"upnp:class"};
        $main::REINDEX_DATA{$artist}{$album}{$title}{"upnp:originalTrackNumber"} = $entry->{"upnp:originalTrackNumber"};
    }

    if ($main::REINDEX{start} >= $result->getValue("TotalMatches")) {
        %main::MUSIC = %main::REINDEX_DATA;
        undef %main::REINDEX_DATA;

        sonos_musicdb_save();
        $main::UPDATEID{ShareIndexInProgress2} = 0;
        sonos_process_waiting("MUSIC");
    } else {
        add_timeout (time(), \&sonos_fetch_music_callback);
    }
}
###############################################################################
sub sonos_musicdb_save {
    delete $main::MUSIC{"#Playlists"};
    delete $main::MUSIC{"#Radio"};
    delete $main::MUSIC{"#LineIn"};
    $main::MUSIC{_info}->{ShareListUpdateID} = $main::UPDATEID{ShareListUpdateID};
    $main::MUSIC{_info}->{version} = $main::VERSION;

    {
        local $Data::Dumper::Purity = 1;
        Log (1, "Saving Music DB");
        open (DB, ">musicdb.pl");
        my $dumper = Data::Dumper->new( [\%main::MUSIC], [ qw( *main::MUSIC) ] );
        print DB $dumper->Dump();
        close DB;
        Log (1, "Finshed Saving Music DB");
    }

    delete $main::MUSIC{_info};

    add_timeout (time(), \&sonos_fetch_radio_stations);
    add_timeout (time(), \&sonos_fetch_playlists);
    add_timeout (time(), \&sonos_fetch_linein);
}
###############################################################################
sub sonos_musicdb_load {
    if ( -f "musicdb.pl") {
        Log (1, "Loading Music DB");
        do "musicdb.pl";
        $main::LASTUPDATE  = $main::SONOS_UPDATENUM;
        $main::MUSICUPDATE = $main::SONOS_UPDATENUM++;
        Log (1, "Finished Loading Music DB");

        # To make life easier, we always update the music index on new version install
        if ($main::MUSIC{_info}->{version} != $main::VERSION) {
            undef $main::MUSIC;
            Log (1, "Old version of Music DB, must rebuild");
        }
        $main::UPDATEID{ShareListUpdateID} = $main::MUSIC{_info}->{ShareListUpdateID};
        delete $main::MUSIC{_info};
    }
}
###############################################################################
sub sonos_prefsdb_save {
    {
        local $Data::Dumper::Purity = 1;
        Log (1, "Saving Prefs DB");
        open (DB, ">prefsdb.pl");
        my $dumper = Data::Dumper->new( [\%main::PREFS], [ qw( *main::PREFS) ] );
        print DB $dumper->Dump();
        close DB;
        Log (1, "Finshed Saving Prefs DB");
    }
}
###############################################################################
sub sonos_prefsdb_load {
    if ( -f "prefsdb.pl") {
        Log (1, "Loading Prefs DB");
        do "prefsdb.pl";
    }
}
###############################################################################
sub sonos_renew_subscriptions {
    foreach my $sub (keys %main::SUBSCRIPTIONS) {
        Log (3, "renew $sub");
        $main::SUBSCRIPTIONS{$sub}->renew;
    }
    add_timeout(time()+$main::RENEW_SUB_TIME, \&sonos_renew_subscriptions);
}

###############################################################################
sub sonos_location_to_id {
    my ($location) = @_;

    foreach my $zone (keys %main::ZONES) {
        return $zone if ($main::ZONES{$zone}->{Location} eq $location);
    }
    return undef;
}

###############################################################################
sub sonos_upnp_update {
    my ($service, %properties) = @_;

    Log (2, "Event received for service=" . $service->{BASE});

    # Save off the zone names
    if ($service->serviceId =~ /serviceId:ZoneGroupTopology/) {
        my $tree = XMLin(decode_entities($properties{ZoneGroupState}), 
                forcearray => ["ZoneGroup", "ZoneGroupMember"]);
        Log(4, "ZoneGroupTopology " . Dumper($tree));
        foreach my $group (@{$tree->{ZoneGroup}}) {
            my %zonegroup = %{$group};

            foreach my $member (@{$zonegroup{ZoneGroupMember}}) {
                my $zkey = $member->{UUID};
                foreach my $mkey (keys %{$member}) {
                    $main::ZONES{$zkey}->{$mkey} = $member->{$mkey};
                }
                $main::ZONES{$zkey}->{Coordinator} = $zonegroup{Coordinator};
            }
        }
        $main::LASTUPDATE  = $main::SONOS_UPDATENUM;
        $main::ZONESUPDATE = $main::SONOS_UPDATENUM++;

        sonos_process_waiting("ZONES");

        return;
    }

    my $zone = sonos_location_to_id($service->{BASE});

    # Save off the current status
    if ($service->serviceId =~ /serviceId:RenderingControl/) {
        my $tree = XMLin(decode_entities($properties{LastChange}), 
                forcearray => ["ZoneGroup"], 
                keyattr=>{"Volume"   => "channel",
                          "Mute"     => "channel",
                          "Loudness" => "channel"});
        Log(4, "RenderingControl " . Dumper($tree));
        foreach my $key ("Volume", "Treble", "Bass", "Mute", "Loudness") {
            if ($tree->{InstanceID}->{$key}) {
                $main::ZONES{$zone}->{RENDER}->{$key} = $tree->{InstanceID}->{$key};
                $main::LASTUPDATE                 = $main::SONOS_UPDATENUM;
                $main::ZONES{$zone}->{RENDER}->{LASTUPDATE} = $main::SONOS_UPDATENUM++;
            }
        }

        sonos_process_waiting("RENDER", $zone);

        return;
    }

    if ($service->serviceId =~ /serviceId:AVTransport/) {
        my $tree = XMLin(decode_entities($properties{LastChange}));
        Log(4, "AVTransport " . Dumper($tree));

        foreach my $key ("CurrentTrackMetaData", "CurrentPlayMode", "NumberOfTracks", "CurrentTrack", "TransportState", "AVTransportURIMetaData", "AVTransportURI", "r:NextTrackMetaData", "CurrentTrackDuration") {
            if ($tree->{InstanceID}->{$key}) {
                $main::LASTUPDATE             = $main::SONOS_UPDATENUM;
                $main::ZONES{$zone}->{AV}->{LASTUPDATE} = $main::SONOS_UPDATENUM++;
                if ($tree->{InstanceID}->{$key}->{val} =~ /^</) {
                    $main::ZONES{$zone}->{AV}->{$key} = \%{XMLin($tree->{InstanceID}->{$key}->{val})};
                } else {
                    $main::ZONES{$zone}->{AV}->{$key} = $tree->{InstanceID}->{$key}->{val};
                }
            }  
        }

        sonos_process_waiting("AV", $zone);
        return;
    }

    if ($service->serviceId =~ /serviceId:ContentDirectory/) {
        Log(4, "ContentDirectory " . Dumper(\%properties));

        if (!defined $main::ZONES{$zone}->{QUEUE} || $properties{ContainerUpdateIDs} =~ /AI:/) {
            Log (1, "Refetching Linein for $main::ZONES{$zone}->{ZoneName} updateid $properties{ContainerUpdateIDs}");
            sonos_fetch_linein($zone);
        }

        if (!defined $main::ZONES{$zone}->{QUEUE} || $properties{ContainerUpdateIDs} =~ /Q:0/) {
            Log (1, "Refetching Q for $main::ZONES{$zone}->{ZoneName} updateid $properties{ContainerUpdateIDs}");
            $main::ZONES{$zone}->{QUEUE} = upnp_content_dir_browse($zone, "Q:0");
            $main::LASTUPDATE = $main::SONOS_UPDATENUM;
            $main::QUEUEUPDATE{$zone} = $main::SONOS_UPDATENUM++;
            sonos_process_waiting("QUEUE", $zone);
        }

        if (defined $properties{ShareIndexInProgress}) {
            $main::UPDATEID{ShareIndexInProgress} = $properties{ShareIndexInProgress};
        }

        if (defined $properties{MasterRadioUpdateID} && 
            (!defined $main::UPDATEID{MasterRadioUpdateID} || $properties{MasterRadioUpdateID} ne $main::UPDATEID{MasterRadioUpdateID})) {
            $main::UPDATEID{MasterRadioUpdateID} = $properties{MasterRadioUpdateID};
            Log (1, "Refetching Radio Stations, update id $properties{MasterRadioUpdateID}");
            add_timeout (time(), \&sonos_fetch_radio_stations);
        }

        if (defined $properties{SavedQueuesUpdateID} && 
            (!defined $main::UPDATEID{SavedQueuesUpdateID} || $properties{SavedQueuesUpdateID} ne $main::UPDATEID{SavedQueuesUpdateID})) {
            $main::UPDATEID{SavedQueuesUpdateID} = $properties{SavedQueuesUpdateID};
            Log (1, "Refetching Playlists, update id $properties{SavedQueuesUpdateID}");
            add_timeout (time(), \&sonos_fetch_playlists);
        }

        if (defined $properties{ShareListUpdateID} && 
            (!defined $main::UPDATEID{ShareListUpdateID} || $properties{ShareListUpdateID} ne $main::UPDATEID{ShareListUpdateID})) {
            $main::UPDATEID{ShareListUpdateID} = $properties{ShareListUpdateID};
            Log (1, "Refetching Index, update id $properties{ShareListUpdateID}");
            add_timeout (time(), \&sonos_fetch_music);
        }
        return;
    }


    while (my ($key, $val) = each %properties) {
        if ($val =~ /&lt/) {
            my $d = decode_entities($val);
            my $tree = XMLin($d, forcearray => ["ZoneGroup"], keyattr=>{"ZoneGroup" => "ID"});
            Log(3, "Property ${key}'s value is " . Dumper($tree));
        } else {
            Log(3, "Property ${key}'s value is " . $val);
        }
    }
}

###############################################################################
sub sonos_music_class {
    my ($mpath) = @_;

    return sonos_music_entry($mpath)->{"upnp:class"};
}
###############################################################################
sub sonos_music_entry {
    my ($mpath) = @_;

    my @parts = split("/", $mpath);

    if ($#parts == 1) {
        return $main::MUSIC{$parts[1]}
    } elsif ($#parts == 2) {
        return $main::MUSIC{$parts[1]}{$parts[2]}
    } elsif ($#parts == 3) {
        return $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]}
    } 

    return undef;
}
###############################################################################
sub sonos_avtransport_set_radio {
    my ($zone, $mpath) = @_;

    my @parts = split("/", $mpath);

    my $entry;
    if (($#parts == 2) && ($parts[1] eq "#Radio")) {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]};
        return;
    } else {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]};
    }

# So, I'm very lazy :-)
    my $urimetadata = '&lt;DIDL-Lite xmlns:dc=&quot;http://purl.org/dc/elements/1.1/&quot; xmlns:upnp=&quot;urn:schemas-upnp-org:metadata-1-0/upnp/&quot; xmlns:r=&quot;urn:schemas-rinconnetworks-com:metadata-1-0/&quot; xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/&quot;&gt;&lt;item id=&quot;' . $entry->{id} . '&quot; parentID=&quot;' . $entry->{parentID} . '&quot; restricted=&quot;true&quot;&gt;&lt;dc:title&gt;'. $entry->{"dc:title"} .  '&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=&quot;cdudn&quot; nameSpace=&quot;urn:schemas-rinconnetworks-com:metadata-1-0/&quot;&gt;RINCON_AssociatedZPUDN&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;';

    upnp_avtransport_set_uri($zone, $entry->{res}->{content}, decode_entities($urimetadata));

    return;
}
###############################################################################
sub sonos_avtransport_set_queue {
    my ($zone) = @_;

    upnp_avtransport_set_uri($zone, "x-rincon-queue:" . $zone . "#0", "");

    return;
}
###############################################################################
sub sonos_avtransport_set_linein {
    my ($zone, $mpath) = @_;

    my @parts = split("/", $mpath);

    my $entry;
    if (($#parts == 2) && ($parts[1] eq "#LineIn")) {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]};
    } else {
        return;
    }

# So, I'm very lazy :-)
    my $urimetadata = '&lt;DIDL-Lite xmlns:dc=&quot;http://purl.org/dc/elements/1.1/&quot; xmlns:upnp=&quot;urn:schemas-upnp-org:metadata-1-0/upnp/&quot; xmlns:r=&quot;urn:schemas-rinconnetworks-com:metadata-1-0/&quot; xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/&quot;&gt;&lt;item id=&quot;AI:0&quot; parentID=&quot;AI:&quot; restricted=&quot;true&quot;&gt;&lt;dc:title&gt;' . $parts[2] .'&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem&lt;/upnp:class&gt;&lt;desc id=&quot;cdudn&quot; nameSpace=&quot;urn:schemas-rinconnetworks-com:metadata-1-0/&quot;&gt;'. $entry->{zone} . '&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;';

    upnp_avtransport_set_uri($zone, $entry->{res}->{content}, decode_entities($urimetadata));

    return;
}
###############################################################################
sub sonos_avtransport_add {
    my ($zone, $mpath) = @_;

    my @parts = split("/", $mpath);

    if ($#parts == 0) {
        foreach my $music (sort {$a cmp $b} keys %main::MUSIC) {
            sonos_avtransport_add($zone, "/$music");
        }
        return;
    } elsif ($#parts == 1) {
        foreach my $music (sort {$a cmp $b} keys %{$main::MUSIC{$parts[1]}}) {
            sonos_avtransport_add($zone, "$mpath/$music");
        }
        return;
    } elsif (($#parts == 2) && ($parts[1] ne "#Playlists") && ($parts[1] ne "#LineIn")) {
        my $mhash = \%{$main::MUSIC{$parts[1]}{$parts[2]}};
        foreach my $music (sort {$mhash->{$a}{"upnp:originalTrackNumber"} <=> $mhash->{$b}{"upnp:originalTrackNumber"}} keys %{$mhash}) {
            sonos_avtransport_add($zone, "$mpath/$music");
        }
        return;
    } 
    
    my $entry;
    if (($#parts == 2) && (($parts[1] eq "#Playlists") || ($parts[1] eq "#LineIn"))) {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]};
    } else {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]};
    }
    if ($entry->{"upnp:class"} eq "object.item.audioItem.audioBroadcast") {
        return
    } 

    upnp_avtransport_add_uri($zone, $entry->{res}->{content}, "");

    return;
}

###############################################################################
sub sonos_add_waiting {
    my $what = shift @_;
    my $zone = shift @_;
    my $cb   = shift @_;

    push @{$main::WAITING{$what}{$zone}}, [$cb, @_];
}
###############################################################################
sub sonos_process_waiting {
    my ($what, $zone) = @_;

    sonos_process_waiting_internal ($what, $zone, $what, $zone) if (defined $zone);
    sonos_process_waiting_internal ($what, "*", $what, $zone);
    sonos_process_waiting_internal ("*", $zone, $what, $zone) if (defined $zone);
    sonos_process_waiting_internal ("*", "*", $what, $zone);
}

###############################################################################
sub sonos_process_waiting_internal {
    my ($mwhat, $mzone, $what, $zone) = @_;

    return if (!defined $main::WAITING{$mwhat}{$mzone});
    my @waiting = @{$main::WAITING{$mwhat}{$mzone}}; 
    @{$main::WAITING{$mwhat}{$mzone}} = ();

    while ($#waiting >= 0) {
        my($callback, @args) = @{shift @waiting};
        &$callback($what, $zone, @args);
    }
}

###############################################################################
sub sonos_link_all_zones {
    my ($masterzone) = @_;

    foreach my $linkedzone (keys %main::ZONES) {
        next if ($linkedzone eq $masterzone);
        upnp_avtransport_set_uri($masterzone, "x-rincon-queue:" . $linkedzone . "#0", "");
    }

}
###############################################################################
sub sonos_link_zone {
    my ($masterzone, $linkedzone) = @_;
    upnp_avtransport_set_uri($masterzone, "x-rincon-queue:" . $linkedzone . "#0", "");
}

###############################################################################
sub sonos_unlink_zone {
    my ($linkedzone) = @_;
    upnp_avtransport_standalone_coordinator($linkedzone);
    upnp_avtransport_set_uri($linkedzone, "x-rincon-queue:" . $linkedzone . "#0", "");

    # Perform the unlink locally also
    $main::ZONES{$linkedzone}->{Coordinator} = $linkedzone;
}

###############################################################################
# UPNP
###############################################################################

###############################################################################
sub upnp_device_get_service {
    my ($device, $name) = @_;

    my $service = $device->getService($name);
    return $service if ($service);

    for my $child ($device->children) {
        $service = $child->getService($name);
        return $service if ($service);
    }
    return undef;
}

###############################################################################
sub upnp_zone_get_service {
    my ($zone, $name) = @_;

    if (! defined $main::ZONES{$zone}->{Location} || ! defined $main::DEVICE{$main::ZONES{$zone}->{Location}}) {
        main::Log(0, "Zone $zone not found");
        return undef;
    }

    return upnp_device_get_service($main::DEVICE{$main::ZONES{$zone}->{Location}}, $name);
}

###############################################################################
sub upnp_content_dir_browse {
    my ($zone, $objectid) = @_;

    my $start = 0;
    my @data = ();
    my $result;

    my $contentDir = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:ContentDirectory:1");
    return undef if (! defined $contentDir);
    my $contentDirProxy = $contentDir->controlProxy; 

    do {
        $result = $contentDirProxy->Browse($objectid, 'BrowseDirectChildren', 
                                           'dc:title,res,dc:creator,upnp:artist,upnp:album', 
                                           $start, 5000, "");

        return undef if (!$result->isSuccessful);

        $start += $result->getValue("NumberReturned");

        my $results = $result->getValue("Result");

        my $tree = XMLin($results, forcearray => ["item", "container"], keyattr=>{"item" => "ID"});

        push(@data, $tree->{item}) if (defined $tree->{item});
        push(@data, $tree->{container}) if (defined $tree->{container});
    } while ($start < $result->getValue("TotalMatches"));

    return \@{$data[0]} if (defined $data[0]);
    return \@data; # This is empty
}

###############################################################################
sub upnp_content_dir_delete {
    my ($zone, $objectid) = @_;

    my $contentDir = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:ContentDirectory:1");
    my $contentDirProxy = $contentDir->controlProxy; 

    $contentDirProxy->DestroyObject($objectid);
}

###############################################################################
sub upnp_content_dir_refresh_share_index {
    my @zones = keys (%main::ZONES);
    my $zone = $main::ZONES{$zones[0]}->{Coordinator};
    my $contentDir = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:ContentDirectory:1");
    my $contentDirProxy = $contentDir->controlProxy; 
    $contentDirProxy->RefreshShareIndex();
}

###############################################################################
sub upnp_avtransport_remove_track {
    my ($zone, $objectid) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->RemoveTrackFromQueue("0", $objectid);
    return;
}
###############################################################################
sub upnp_avtransport_play {
    my ($zone) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->Play("0", "1");
    return $result;
}
###############################################################################
sub upnp_avtransport_seek {
    my ($zone,$queue) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    $queue =~ s,^.*/,,;

    my $result = $avTransportProxy->Seek("0", "TRACK_NR", $queue);
    return $result;
}
###############################################################################
sub upnp_render_mute {
    my ($zone,$on) = @_;

    if (!defined $on) {
        return $main::ZONES{$zone}->{RENDER}->{Mute}->{Master}->{val};
    }

    my $render = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:RenderingControl:1");
    my $renderProxy = $render->controlProxy;
    my $result = $renderProxy->SetMute("0", "Master", $on);
    $main::ZONES{$zone}->{RENDER}->{Mute}->{Master}->{val} = $on;
    return $on;
}
###############################################################################
sub upnp_render_volume_change {
    my ($zone,$change) = @_;

    my $vol = $main::ZONES{$zone}->{RENDER}->{Volume}->{Master}->{val} + $change;
    return upnp_render_volume($zone,$vol);
}
###############################################################################
sub upnp_render_volume {
    my ($zone,$vol) = @_;

    if (!defined $vol) {
        return $main::ZONES{$zone}->{RENDER}->{Volume}->{Master}->{val};
    }

    $vol = 100 if ($vol > 100);
    $vol = 0 if ($vol < 0);

    my $render = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:RenderingControl:1");
    my $renderProxy = $render->controlProxy;
    my $result = $renderProxy->SetVolume("0", "Master", $vol);
    $main::ZONES{$zone}->{RENDER}->{Volume}->{Master}->{val} = $vol;
    return $vol;
}
###############################################################################
sub upnp_avtransport_action {
    my ($zone,$action) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->$action("0");
    return $result;
}
###############################################################################
sub upnp_avtransport_repeat {
    my ($zone, $repeat) = @_;

    my $str = $main::ZONES{$zone}->{AV}->{CurrentPlayMode};

    if (!defined $repeat) {
        return 0 if ($str eq "NORMAL" || $str eq "SHUFFLE");
        return 1;
    }

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    if ($str eq "NORMAL") {
        $str = "REPEAT_ALL" if ($repeat);
    } elsif ($str eq "REPEAT_ALL") {
        $str = "NORMAL" if (!$repeat);
    } elsif ($str eq "SHUFFLE_NOREPEAT") {
        $str = "SHUFFLE" if ($repeat);
    } elsif ($str eq "SHUFFLE") {
        $str = "SHUFFLE_NOREPEAT" if (!$repeat);
    }
    my $result = $avTransportProxy->SetPlayMode("0", $str);
    $main::ZONES{$zone}->{AV}->{CurrentPlayMode} = $str;
    return $repeat;
}
###############################################################################
sub upnp_avtransport_shuffle {
    my ($zone, $shuffle) = @_;

    my $str = $main::ZONES{$zone}->{AV}->{CurrentPlayMode};

    if (!defined $shuffle) {
        return 0 if ($str eq "NORMAL" || $str eq "REPEAT_ALL");
        return 1;
    }

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    if ($str eq "NORMAL") {
        $str = "SHUFFLE_NOREPEAT" if ($shuffle);
    } elsif ($str eq "REPEAT_ALL") {
        $str = "SHUFFLE" if ($shuffle);
    } elsif ($str eq "SHUFFLE_NOREPEAT") {
        $str = "NORMAL" if (!$shuffle);
    } elsif ($str eq "SHUFFLE") {
        $str = "REPEAT_ALL" if (!$shuffle);
    }

    my $result = $avTransportProxy->SetPlayMode("0", $str);
    $main::ZONES{$zone}->{AV}->{CurrentPlayMode} = $str;
    return $shuffle;
}
###############################################################################
sub upnp_avtransport_set_uri {
    my ($zone, $uri, $metadata) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $result = $avTransportProxy->SetAVTransportURI(0, $uri, $metadata);
}

###############################################################################
sub upnp_avtransport_add_uri {
    my ($zone, $uri, $metadata) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $result = $avTransportProxy->AddURIToQueue(0, $uri, $metadata);
}

###############################################################################
sub upnp_avtransport_standalone_coordinator {
    my ($zone) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $result = $avTransportProxy->BecomeCoordinatorOfStandaloneGroup(0);
}

###############################################################################
sub upnp_avtransport_save {
    my ($zone, $name) = @_;

    my $avTransport = upnp_zone_get_service($zone, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $result = $avTransportProxy->SaveQueue(0, $name, "");
}
###############################################################################
sub upnp_search_cb {
    my ($search, $device, $action) = @_;
    if ($action eq 'deviceAdded') {
        Log(1, "Device: " . $device->friendlyName . " added.");
        #print Dumper($device);

        $main::DEVICE{$device->{LOCATION}} = $device;
         
        my $zonegroup = upnp_device_get_service($device, "urn:schemas-upnp-org:service:ZoneGroupTopology:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},zonegroup"} = $zonegroup->subscribe(\&sonos_upnp_update);

        my $contentDir = upnp_device_get_service($device, "urn:schemas-upnp-org:service:ContentDirectory:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},contentDir"} = $contentDir->subscribe(\&sonos_upnp_update);

        my $avtransport = upnp_device_get_service($device, "urn:schemas-upnp-org:service:AVTransport:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},avtransport"} = $avtransport->subscribe(\&sonos_upnp_update);

        my $rendercontrol = upnp_device_get_service($device, "urn:schemas-upnp-org:service:RenderingControl:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},rendercontrol"} = $rendercontrol->subscribe(\&sonos_upnp_update);

        my $deviceprops = upnp_device_get_service($device, "urn:schemas-upnp-org:service:DeviceProperties:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},deviceprops"} = $deviceprops->subscribe(\&sonos_upnp_update);
    }
    elsif ($action eq 'deviceRemoved') {
        Log(1, "Device: " . $device->friendlyName . " removed.");
    } else {
        Log(1, "Device: " . $device->friendlyName . " unknown action: $action.");
    }
}
###############################################################################
# HTTP
###############################################################################

###############################################################################
sub http_register_handler  {
    my ($path, $callback) = @_;

    $main::HTTP_HANDLERS{$path} = $callback;
}

###############################################################################
sub http_albumart_request {
    my ($c, $r) = @_;

    my $uri = $r->uri;
    my %qf = $uri->query_form;

    my @zones = keys (%main::ZONES);
    my $zone = $main::ZONES{$zones[0]}->{Coordinator};
    my $location = $main::ZONES{$zone}->{Location};
    my @ip = split(/\//, $location);

    # Get the album art given the mpath instead of the albumart path
    if ($uri->path eq "/getAA") {
        my $mpath = decode("UTF-8", $qf{mpath});
        my @parts = split("/", $mpath);

        if ($#parts == 2) {
            $uri = "/getaa?uri=" . (values %{$main::MUSIC{$parts[1]}{$parts[2]}} ) [0]->{res}->{content};
        } elsif ($#parts == 3) {
            $uri = "/getaa?uri=" . $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]}->{res}->{content};
        } else {
            $c->send_error(HTTP::Status::RC_NOT_FOUND);
            $c->force_last_request;
        }
    }

    my $request = "http://$ip[2]" . $uri;
    my $response = $main::useragent->get($request);

    # Cache album art
    $response->header( "Cache-Control" => "max-age=" . $main::AACACHE );
    $response->header( "Expires" => time() + $main::AACACHE );

    $c->send_response($response);
    $c->force_last_request;
}
###############################################################################
sub http_base_url {
    my ($r) = @_;

    my $baseurl;

    if ($r->header("host")) {
        $baseurl = "http://".$r->header("host");
    } else {
        $baseurl = "http://". UPnP::Common::LOCAL_IP.":".$main::HTTP_PORT;
    }
}
###############################################################################
sub http_handle_request {
    my ($c, $r) = @_;

    # No r, just return
    if (!$r || !$r->uri) {
        Log (1, "Missing Request");
        return;
    }

    my $uri = $r->uri;

    my $path = $uri->path;
    my $baseurl = http_base_url($r);

    if (($path eq "/") || ($path =~ /\.\./)) {
        $c->send_redirect("$baseurl/$main::DEFAULT");
        $c->force_last_request;
        return;
    }

    my %qf = $uri->query_form;
    Log (1, "URI: $uri");

    if ($main::HTTP_HANDLERS{$path}) {
        my $callback = $main::HTTP_HANDLERS{$path};
        &$callback($c, $r);
        return;
    }

    # Find where on disk
    my $diskpath;
    my $tmplhook;
    if (-e "html/$path") {
        $diskpath = "html$path";
    } else {
        my @parts = split("/", $path);
        my $plugin = $parts[1];
        splice(@parts, 0, 2);
        my $restpath = join("/", @parts);
        if ($main::PLUGINS{$plugin} && $main::PLUGINS{$plugin}->{html} && 
            -e $main::PLUGINS{$plugin}->{html} . $restpath) {
            $diskpath = $main::PLUGINS{$plugin}->{html} . $restpath;
            $tmplhook = $main::PLUGINS{$plugin}->{tmplhook};
        }
    }

    # File doesn't exist
    if ( ! $diskpath) {
        $c->send_error(HTTP::Status::RC_NOT_FOUND);
        $c->force_last_request;
        return;
    }

    # File is a directory, redirect for the browser
    if (-d $diskpath) {
        if ($path =~ /\/$/) {
            $c->send_redirect($baseurl . $path . "index.html");
        } else {
            $c->send_redirect($baseurl . $path . "/index.html");
        }
        $c->force_last_request;
        return;
    }

    # File isn't HTML/XML, just send it back raw
    if (!($path =~ /\.html/) && !($path =~ /\.xml/)) {
        $c->send_file_response($diskpath);
        $c->force_last_request;
        return;
    }

    # 0 - not handled, 1 - handled send reply, >= 2 - handled routine will send reply
    my $response = 0;

    if ($qf{action}) {
        if (defined $qf{zone} && $qf{zone} ne "") {
            $response = http_handle_zone_action($c, $r, $path);
        }
        if (! $response) {
            $response = http_handle_action($c, $r, $path);
        }
    }

    if ($qf{NoWait}) {
        $response = 1;
    }

    if ($response == 2) {
        sonos_add_waiting("AV", "*", \&http_send_response, $c, $r, $diskpath, $tmplhook);
    } elsif ($response == 3) {
        sonos_add_waiting("RENDER", "*", \&http_send_response, $c, $r, $diskpath, $tmplhook);
    } elsif ($response == 4) {
        sonos_add_waiting("*", "*", \&http_send_response, $c, $r, $diskpath, $tmplhook);
    } else {
        http_send_response("*", "*", $c, $r, $diskpath, $tmplhook);
    }
}

###############################################################################
sub http_handle_zone_action {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;
    my $mpath = decode("UTF-8", $qf{mpath});

    my $zone = $qf{zone};
    if ($qf{action} eq "Remove") {
        upnp_avtransport_remove_track($zone, $qf{queue});
        return 2;
    } elsif ($qf{action} eq "RemoveAll") {
        upnp_avtransport_action($zone, "RemoveAllTracksFromQueue");
        return 2;
    } elsif ($qf{action} eq "Play") {
        upnp_avtransport_play($zone);
        return 2;
    } elsif ($qf{action} =~ /(Next|Previous|Stop|Pause)/) {
        upnp_avtransport_action($zone, $qf{action});
        return 2;
    } elsif ($qf{action} eq "ShuffleOn") {
        upnp_avtransport_shuffle($zone, 1);
        return 2;
    } elsif ($qf{action} eq "ShuffleOff") {
        upnp_avtransport_shuffle($zone, 0);
        return 2;
    } elsif ($qf{action} eq "RepeatOn") {
        upnp_avtransport_repeat($zone, 1);
        return 2;
    } elsif ($qf{action} eq "RepeatOff") {
        upnp_avtransport_repeat($zone, 0);
        return 2;
    } elsif ($qf{action} eq "Seek") {
        if (! ($main::ZONES{$zone}->{AV}->{AVTransportURI} =~ /queue/)) {
            sonos_avtransport_set_queue($zone);
        }
        upnp_avtransport_seek($zone, $qf{queue});
        return 2;
    } elsif ($qf{action} eq "MuteOn") {
        upnp_render_mute($zone, 1);
        return 3;
    } elsif ($qf{action} eq "MuteOff") {
        upnp_render_mute($zone, 0);
        return 3;
    } elsif ($qf{action} eq "MuchSofter") {
        upnp_render_volume_change($zone, -5);
        return 3;
    } elsif ($qf{action} eq "Softer") {
        upnp_render_volume_change($zone, -1);
        return 3;
    } elsif ($qf{action} eq "Louder") {
        upnp_render_volume_change($zone, +1);
        return 3;
    } elsif ($qf{action} eq "MuchLouder") {
        upnp_render_volume_change($zone, +5);
        return 3;
    } elsif ($qf{action} eq "SetVolume") {
        upnp_render_volume($zone, $qf{volume});
        return 3;
    } elsif ($qf{action} eq "Save") {
        upnp_avtransport_save($zone, $qf{savename});
        return 0;
    } elsif ($qf{action} eq "AddMusic") {
        my $class = sonos_music_class($mpath);
        if ($class eq "object.item.audioItem.audioBroadcast") {
            sonos_avtransport_set_radio($zone, $mpath);
            return 2;
        } elsif ($class eq "object.item.audioItem") {
            sonos_avtransport_set_linein($zone, $mpath);
        } else {
            sonos_avtransport_add($zone, $mpath);
            return 4;
        }
    } elsif ($qf{action} eq "DeleteMusic") {
        my @parts = split("/", $mpath);
        if ($#parts == 2 && sonos_music_class($mpath) eq "object.container.playlist") {
            my $id = $main::MUSIC{$parts[1]}{$parts[2]}{"id"};
            upnp_content_dir_delete($zone, $id);
        }
        return 0;
    } elsif ($qf{action} eq "PlayMusic") {
        my $class = sonos_music_class($mpath);
        if ($class eq "object.item.audioItem.audioBroadcast") {
            sonos_avtransport_set_radio($zone, $mpath);
        } elsif ($class eq "object.item.audioItem") {
            sonos_avtransport_set_linein($zone, $mpath);
        } else {
            if (! ($main::ZONES{$zone}->{AV}->{AVTransportURI} =~ /queue/)) {
                sonos_avtransport_set_queue($zone);
            }
            upnp_avtransport_action($zone, "RemoveAllTracksFromQueue");
            sonos_avtransport_add($zone, $mpath);
            upnp_avtransport_play($zone);
        }

        return 2;
    } else {
        return 0;
    }
    return 1;
}

###############################################################################
sub http_handle_action {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    if ($qf{action} eq "ReIndex") {
        sonos_reindex();
    } elsif ($qf{action} eq "Wait" && $qf{lastupdate}) {
        if ($main::LASTUPDATE > $qf{lastupdate}) {
            return 1;
        } else {
            return 4;
        }
    } else {
        return 0;
    }
    return 1;
}
###############################################################################
sub http_build_zone_data {
my ($zone, $updatenum) = @_;

    my %activedata;

    $activedata{"ACTIVE_ZONE"}      = encode_entities($main::ZONES{$zone}->{ZoneName});
    $activedata{"ACTIVE_ZONEID"}    = uri_escape($zone); 
    $activedata{"ACTIVE_VOLUME"}    = $main::ZONES{$zone}->{RENDER}->{Volume}->{Master}->{val};

    my $lastupdate;
    if ($main::ZONES{$zone}->{RENDER}->{LASTUPDATE} > $main::ZONES{$zone}->{AV}->{LASTUPDATE}) {
        $lastupdate = $main::ZONES{$zone}->{RENDER}->{LASTUPDATE};
    } else {
        $lastupdate = $main::ZONES{$zone}->{AV}->{LASTUPDATE};
    }

    $activedata{"ACTIVE_LASTUPDATE"} = $lastupdate;
    $activedata{"ACTIVE_UPDATED"}    = ($lastupdate > $updatenum);

    if ($main::ZONES{$zone}->{RENDER}->{Mute}->{Master}->{val}) {
        $activedata{"ACTIVE_MUTED"} = 1;
    } else {
        $activedata{"ACTIVE_MUTED"} = 0;
    }

    my $curtrack     = $main::ZONES{$zone}->{AV}->{CurrentTrackMetaData};
    my $curtransport = $main::ZONES{$zone}->{AV}->{AVTransportURIMetaData};

    $activedata{"ACTIVE_NAME"}      = "";
    $activedata{"ACTIVE_ARTIST"}    = "";
    $activedata{"ACTIVE_ALBUM"}     = "";
    $activedata{"ACTIVE_ISSONG"}    = 1;
    $activedata{"ACTIVE_ISRADIO"}   = 0;
    $activedata{"ACTIVE_TRACK_NUM"} = 0;
    $activedata{"ACTIVE_TRACK_TOT"} = 0;
    $activedata{"ACTIVE_MODE"}      = 0;
    $activedata{"ACTIVE_PAUSED"}    = 0;
    $activedata{"ACTIVE_STOPPED"}   = 0;
    $activedata{"ACTIVE_PLAYING"}   = 0;
    $activedata{"ACTIVE_SHUFFLE"}   = 0;
    $activedata{"ACTIVE_REPEAT"}    = 0;
    $activedata{"ACTIVE_POSITION"}  = 0;
    $activedata{"ACTIVE_LENGTH"}    = 0;
    $activedata{"ACTIVE_ALBUMART"}  = "";
    $activedata{"ACTIVE_CONTENT"}   = "";

    if ($curtrack) {
        if ($curtrack->{item}->{res}{content}) {
            $activedata{"ACTIVE_CONTENT"}   = $curtrack->{item}->{res}{content};
        }

        if ($curtrack->{item}->{"upnp:albumArtURI"}) {
            $activedata{"ACTIVE_ALBUMART"}  = $curtrack->{item}->{"upnp:albumArtURI"};
        }

        if ($curtransport && $curtransport->{item}->{"upnp:class"} eq "object.item.audioItem.audioBroadcast") {
            if (!ref($curtrack->{item}->{"r:streamContent"})) {
                $activedata{"ACTIVE_NAME"}  = encode_entities($curtrack->{item}->{"r:streamContent"});
            }

            $activedata{"ACTIVE_ALBUM"}     = encode_entities($curtransport->{item}->{"dc:title"});
            $activedata{"ACTIVE_ISSONG"}    = 0;
            $activedata{"ACTIVE_ISRADIO"}   = 1;
            $activedata{"ACTIVE_TRACK_NUM"} = 0;
            $activedata{"ACTIVE_TRACK_TOT"} = 0;
        } else {

            $activedata{"ACTIVE_NAME"}      = encode_entities($curtrack->{item}->{"dc:title"});
            $activedata{"ACTIVE_ARTIST"}    = encode_entities($curtrack->{item}->{"dc:creator"});
            $activedata{"ACTIVE_ALBUM"}     = encode_entities($curtrack->{item}->{"upnp:album"});
            $activedata{"ACTIVE_TRACK_NUM"} = $main::ZONES{$zone}->{AV}->{CurrentTrack};
            $activedata{"ACTIVE_TRACK_TOT"} = $main::ZONES{$zone}->{AV}->{NumberOfTracks};
        }

        if ($main::ZONES{$zone}->{AV}->{TransportState} eq "PAUSED_PLAYBACK") {
            $activedata{"ACTIVE_MODE"}      = 2;
            $activedata{"ACTIVE_PAUSED"}    = 1;
        } elsif ($main::ZONES{$zone}->{AV}->{TransportState} eq "STOPPED"){
            $activedata{"ACTIVE_MODE"}      = 0;
            $activedata{"ACTIVE_STOPPED"}   = 1;
        } else {
            $activedata{"ACTIVE_MODE"}      = 1;
            $activedata{"ACTIVE_PLAYING"}   = 1;
        }

        if ($main::ZONES{$zone}->{AV}->{CurrentPlayMode} eq "NORMAL") {
        } elsif ($main::ZONES{$zone}->{AV}->{CurrentPlayMode} eq "REPEAT_ALL") {
            $activedata{"ACTIVE_REPEAT"}    = 1;
        } elsif ($main::ZONES{$zone}->{AV}->{CurrentPlayMode} eq "SHUFFLE_NOREPEAT") {
            $activedata{"ACTIVE_SHUFFLE"}   = 1;
        } elsif ($main::ZONES{$zone}->{AV}->{CurrentPlayMode} eq "SHUFFLE") {
            $activedata{"ACTIVE_SHUFFLE"}   = 1;
            $activedata{"ACTIVE_REPEAT"}    = 1;
        }
    }

    if ($main::ZONES{$zone}->{AV}->{CurrentTrackDuration}) {
        my @parts = split(":", $main::ZONES{$zone}->{AV}->{CurrentTrackDuration});
        $activedata{"ACTIVE_LENGTH"}    = $parts[0]*3600+$parts[1]*60+$parts[2];
    }

    my $nexttrack = $main::ZONES{$zone}->{AV}->{"r:NextTrackMetaData"};
    if ($nexttrack) {
        $activedata{"NEXT_NAME"}      = encode_entities($nexttrack->{item}->{"dc:title"});
        $activedata{"NEXT_ARTIST"}    = encode_entities($nexttrack->{item}->{"dc:creator"});
        $activedata{"NEXT_ALBUM"}     = encode_entities($nexttrack->{item}->{"upnp:album"});
        $activedata{"NEXT_ISSONG"}    = 1;
    }

    return %activedata;
}

###############################################################################
sub http_build_queue_data {
my ($zone, $updatenum) = @_;

    my %queuedata;

    $queuedata{QUEUE_ZONE}       = encode_entities($main::ZONES{$zone}->{ZoneName});
    $queuedata{QUEUE_ZONEID}     = uri_escape($zone);
    $queuedata{QUEUE_LASTUPDATE} = $main::QUEUEUPDATE{$zone};
    $queuedata{QUEUE_UPDATED}    = ($main::QUEUEUPDATE{$zone} > $updatenum);

    return %queuedata if (!$main::ZONES{$zone}->{QUEUE} || ! (@{$main::ZONES{$zone}->{QUEUE}}));

    my @loop_data = ();
    foreach my $queue (@{@{$main::ZONES{$zone}->{QUEUE}}}) {
        my %row_data;
        if ($main::ZONES{$zone}->{AV}->{CurrentTrackMetaData} ne "" && 
            $queue->{res}->{content} eq $main::ZONES{$zone}->{AV}->{CurrentTrackMetaData}->{item}->{res}->{content}) {
            if ($main::ZONES{$zone}->{AV}->{TransportState} eq "PLAYING") {
                $row_data{QUEUE_IMG} = "/playing.gif";
            } else {
                $row_data{QUEUE_IMG} = "/paused.gif";
            }
        } else {
            $row_data{QUEUE_IMG} = "/blank.gif";
        }
        $row_data{QUEUE_NAME} = encode_entities($queue->{"dc:title"});
        $row_data{QUEUE_ARG}  = "zone=" . uri_escape($zone). "&queue=$queue->{id}";
        $row_data{QUEUE_ID}   = $queue->{id};
        push(@loop_data, \%row_data);
    }
    $queuedata{QUEUE_LOOP}       = \@loop_data;

    return %queuedata;
}

###############################################################################
# Sort items by coordinators first, for linked zones sort under their coordinator
sub http_zone_sort () {
    my $c = $main::ZONES{$main::ZONES{$main::a}->{Coordinator}}->{ZoneName} cmp 
            $main::ZONES{$main::ZONES{$main::b}->{Coordinator}}->{ZoneName};
    Log(0, "a $main::a b $main::b c $c");
    return $c if ($c != 0);
    return -1 if ($main::ZONES{$main::a}->{Coordinator} eq $main::a);
    return 1 if ($main::ZONES{$main::b}->{Coordinator} eq $main::b);
    return $main::ZONES{$main::a}->{ZoneName} cmp $main::ZONES{$main::b}->{ZoneName};
}
###############################################################################
sub http_send_response {
my ($what, $thezone, $c, $r, $diskpath, $tmplhook) = @_;

    my %qf = $r->uri->query_form;

    my $updatenum = 0;

    $updatenum = $qf{lastupdate} if ($qf{lastupdate});

    # One of ours templates, now fill in the parts we know
    my $template = HTML::Template->new(filename => $diskpath,
                                       die_on_bad_params => 0,
                                       cache => 1,
                                       loop_context_vars => 1);

    $template->param("VERSION"     => $main::VERSION);
    $template->param("LAST_UPDATE" => $main::LASTUPDATE);

    if ($r->header("host")) {
        $template->param("BASE_URL" => "http://".$r->header("host"));
    } else {
        $template->param("BASE_URL" => "http://". UPnP::Common::LOCAL_IP.":".$main::HTTP_PORT);
    }

    # There is a ZONES_LOOP so gather all the information about the zones
    if ($template->query(name => "ZONES_LOOP")) {
        $template->param("ZONES_LASTUPDATE" => $main::ZONESUPDATE);
        $template->param("ZONES_UPDATED" => ($main::ZONESUPDATE > $updatenum));
        my @loop_data = ();
        foreach my $zone (sort http_zone_sort (keys %main::ZONES)) {
            my %row_data = http_build_zone_data($zone, $updatenum);
            $row_data{ZONE_ACTIVE} = (defined $qf{zone} && $zone eq $qf{zone});
            $row_data{ZONE_NAME} = encode_entities($main::ZONES{$zone}->{ZoneName});
            $row_data{ZONE_ARG}  = "zone=".uri_escape($zone);
            if ($main::ZONES{$zone}->{Coordinator} eq $zone) {
                $row_data{ZONE_LINKED}    = 0;
                $row_data{ZONE_LINK}      = "";
                $row_data{ZONE_LINK_NAME} = "";
            } else {
                $row_data{ZONE_LINKED}    = 1;
                $row_data{ZONE_LINK}      = $main::ZONES{$zone}->{Coordinator};
                $row_data{ZONE_LINK_NAME} = $main::ZONES{$main::ZONES{$zone}->{Coordinator}}->{ZoneName};
            }
            push(@loop_data, \%row_data);
        }
        $template->param(ZONES_LOOP => \@loop_data);
    }
    
    # There was a zone param, get the data about single zone
    if (defined $qf{zone} && $qf{zone} ne "") {
        my %activedata = http_build_zone_data($qf{zone}, $updatenum);
        $template->param(%activedata);
    }

    my $all_arg = "";
    $all_arg = $r->uri->query if ($r->uri->query);
    $all_arg =~ s/[&]*action=[^&]+//;
    $all_arg =~ s/[&]*rand=[^&]+//;
    $all_arg =~ s/[&]*mpath=[^&]+//;
    $template->param("ALL_ARG" => "$all_arg&rand=$$.$main::HTTP_COUNTER");
    $main::HTTP_COUNTER++;

    if ($template->query(name => "ALL_QUEUE_LOOP")) {
        my @loop_data = ();
        foreach my $zone (keys %main::ZONES) {
            my %row_data = http_build_queue_data($zone, $updatenum);

            push(@loop_data, \%row_data);
        }
        $template->param(ALL_QUEUE_LOOP => \@loop_data);
    }

    if ($template->query(name => "QUEUE_LOOP") && 
        defined $qf{zone} &&
        $main::ZONES{$qf{zone}}->{QUEUE} && 
        @{$main::ZONES{$qf{zone}}->{QUEUE}}) {

        my %queuedata = http_build_queue_data($qf{zone}, $updatenum);
        $template->param(%queuedata);
    }

    if ($template->query(name => "MUSIC_LOOP") && ((defined $qf{zone} && $main::ZONES{$qf{zone}}->{QUEUE}) || $qf{lastupdate})) {
        my @loop_data = ();

        my $sortsub = sub {};

        if ($qf{msearch}) {

            my $maxsearch = $main::MAX_SEARCH;
            $maxsearch = $qf{maxsearch} if ($qf{maxsearch});
            my $msearch = $qf{msearch};
            my $searchartist = my $searchalbum = my $searchsong = 1;
            if ($msearch =~ /^artist:/) {
                $searchalbum = $searchsong = 0;
                $msearch = substr($msearch, 7); 
            } elsif ($msearch =~ /^album:/) {
                $searchartist = $searchsong = 0;
                $msearch = substr($msearch, 6); 
            } elsif ($msearch =~ /^song:/) {
                $searchartist = $searchalbum = 0;
                $msearch = substr($msearch, 5); 
            }

# Check that RE compiles
            eval "if ('foo' =~ /$msearch/i) {1;}";
            if ($@) {
                Log (2, "Eval failed: $@");
                $msearch = "^\$"; # Match nothing
            }

            foreach my $artist (sort {$a cmp $b} keys %main::MUSIC) {

                if ($searchartist && ($artist =~ /$msearch/i)) {
                    my %row_data;
                    $row_data{MUSIC_NAME} = encode_entities($artist);
                    $row_data{MUSIC_PATH} = "/" . uri_escape_utf8($artist, "^A-Za-z0-9");
                    if (defined $qf{zone}) {
                        $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                              "&mpath=" . $row_data{MUSIC_PATH};
                    }
                    $row_data{MUSIC_ISSONG} = 0;
                    push(@loop_data, \%row_data);
                }

                if ($#loop_data > $maxsearch) {
                    $template->param("MUSIC_ERROR" => "More then $maxsearch matching items.<BR>");
                    last;
                }
                foreach my $album (sort {$a cmp $b} keys %{$main::MUSIC{$artist}}) {
                    if ($searchalbum && ($album =~ /$msearch/i)) {
                        my %row_data;
                        $row_data{MUSIC_NAME} = encode_entities($album);
                        $row_data{MUSIC_PATH} = "/" . uri_escape_utf8($artist, "^A-Za-z0-9") .
                                                "/" . uri_escape_utf8($album, "^A-Za-z0-9") ;
                        if (defined $qf{zone}) {
                            $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                                           "&mpath=" . $row_data{MUSIC_PATH};
                        }
                        if (($artist eq "#Playlists") || ($artist eq "#LineIn")) {
                            $row_data{MUSIC_ISSONG} = 1;
                        } else {
                            $row_data{MUSIC_ISSONG} = 0;
                        }
                        push(@loop_data, \%row_data);
                    }
                    if ($#loop_data > $maxsearch) {
                        $template->param("MUSIC_ERROR" => "More then $maxsearch matching items.<BR>");
                        last;
                    }
                    foreach my $song (sort {$a cmp $b} keys %{$main::MUSIC{$artist}{$album}}) {
                        if ($searchsong && ($song =~ /$msearch/i)) {
                            my %row_data;
                            $row_data{MUSIC_NAME} = encode_entities($song);
                            $row_data{MUSIC_PATH} = "/" . uri_escape_utf8($artist, "^A-Za-z0-9") .
                                                    "/" . uri_escape_utf8($album, "^A-Za-z0-9") .
                                                    "/" . uri_escape_utf8($song, "^A-Za-z0-9");
                            if (defined $qf{zone}) {
                                $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                                      "&mpath=".$row_data{MUSIC_PATH};
                            }
                            $row_data{MUSIC_ISSONG} = 1;
                            push(@loop_data, \%row_data);
                        }
                    }
                }
            }
            $template->param("MUSIC_LOOP" => \@loop_data);
            $template->param("MUSIC_PATH" => "Search: ". uri_escape($qf{msearch}));
            $template->param("MUSIC_SEARCH" => $qf{msearch});
            $template->param("MUSIC_UPDATED" => 1);
            $template->param("MUSIC_LASTUPDATE" => $main::MUSICUPDATE);

        } else {
            my $mpath = "";
            $mpath = join("&",decode("UTF-8", $qf{mpath})) if (defined $qf{mpath});

            my $mhash;
            my @parts = split("/", $mpath);

            if (($#parts > 1) && ($parts[1] eq "#Playlists")) {
                @parts = ("", "#Playlists");
                $mpath = "";
            }

            if (($#parts > 1) && ($parts[1] eq "#LineIn")) {
                @parts = ("", "#LineIn");
                $mpath = "";
            }

            if ($#parts == 1) {
                $mhash = \%{$main::MUSIC{$parts[1]}};
                $sortsub = sub {$a cmp $b};
                $template->param("MUSIC_UPDATED" => 1);
                $mpath = "/".uri_escape_utf8($parts[1], "^A-Za-z0-9");
            } elsif ($#parts == 2) {
                $mhash = \%{$main::MUSIC{$parts[1]}{$parts[2]}};
                $sortsub = sub {$mhash->{$a}{"upnp:originalTrackNumber"} <=> $mhash->{$b}{"upnp:originalTrackNumber"}};
                $template->param("MUSIC_UPDATED" => 1);
                $mpath = "/".uri_escape_utf8( $parts[1], "^A-Za-z0-9").
                         "/".uri_escape_utf8($parts[2], "^A-Za-z0-9");
            } else {
                $template->param("MUSIC_ROOT" => "1");
                $sortsub = sub {$a cmp $b};
                $mhash = \%main::MUSIC;
                $mpath = "";
                $template->param("MUSIC_UPDATED" => (!$qf{NoWait} && ($main::MUSICUPDATE > $updatenum)));
            }
            $template->param("MUSIC_LASTUPDATE" => $main::MUSICUPDATE);
            $template->param("MUSIC_PATH" => $mpath);

            foreach my $music (sort $sortsub keys %$mhash) {
                my %row_data;
                $row_data{MUSIC_NAME} = encode_entities($music);
                $row_data{MUSIC_PATH} = $mpath . "/" . uri_escape_utf8($music, "^A-Za-z0-9");
                if (defined $qf{zone}) {
                    $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                         "&mpath=" . $row_data{MUSIC_PATH};
                }
                if ($#parts == 2 || (($#parts == 1) && (($parts[1] eq "#Playlists") || ($parts[1] eq "#LineIn")))) {
                    $row_data{MUSIC_ISSONG} = 1;
                } else {
                    $row_data{MUSIC_ISSONG} = 0;
                }
                push(@loop_data, \%row_data);
            }
            $template->param("MUSIC_LOOP" => \@loop_data);
        }
    }

    if ($template->query(name => "PLUGIN_LOOP")) {
        my @loop_data = ();

        foreach my $plugin (sort (keys %main::PLUGINS)) {
            next if (! $main::PLUGINS{$plugin}->{link});
            my %row_data;
            $row_data{PLUGIN_LINK} = $main::PLUGINS{$plugin}->{link};
            $row_data{PLUGIN_NAME} = $main::PLUGINS{$plugin}->{name};
            
            push(@loop_data, \%row_data);
        }

        $template->param("PLUGIN_LOOP" => \@loop_data);
    }

    &$tmplhook($c, $r, $diskpath, $template) if ($tmplhook);
    
    my $response = HTTP::Response->new(200, undef, [Connection => "close", "Content-Type" => "text/html; charset=ISO-8859-1"], $template->output);
    $c->send_response($response);
    $c->force_last_request;
    $c->close;
}

###############################################################################
sub http_quit {
    $main::daemon->close();
}

###############################################################################
# UTILS
###############################################################################
sub add_timeout {
    my $time = shift @_;
    my $cb = shift @_;

    @main::TIMERS = sort { $a->[0] <=> $b->[0] } @main::TIMERS, [$time, $cb, @_];
}
###############################################################################
sub Log {
    my $level = shift;
    my $now = strftime "%e/%H%M%S", localtime;
    my @Level = ("crit", "alert", "notice", "info", "debug");

    return if ($level > $main::MAX_LOG_LEVEL);

    my ($package, $filename, $line) = caller(0);
    my ($x, $y, $z, $subroutine)    = caller(1);
    $subroutine = "*unknown*" if (!defined $subroutine);
    print "$now $Level[$level]: $filename:$line $subroutine @_\n";
}
###############################################################################
# Copied from newer URI::Escape module
sub uri_escape_utf8
{
    my $text = shift;
    if ($] < 5.008) {
        $text =~ s/([^\0-\x7F])/do {my $o = ord($1); sprintf("%c%c", 0xc0 | ($o >> 6), 0x80 | ($o & 0x3f)) }/ge;
    }
    else {
        utf8::encode($text);
    }

    return uri_escape($text, @_);
}
